// Aatie.cpp
//

#include <windows.h>
#include <tchar.h>
#include <stdio.h>
#include <d2d1.h>
#include <dwrite.h>
#include <wincodec.h>
#include <wincodecsdk.h>
#include <shlwapi.h>
#include <OCIdl.h>
#include <atlbase.h>
#include <detours.h>

#include "brute_cast.h"

const CLSID CLSID_Aatie = {0xecc37558, 0x63c9, 0x49eb,
                           0xaa, 0x3d, 0x5d, 0x19, 0xf6, 0x99, 0x73, 0x12};
#define CLSID_Aatie_s _T("{ecc37558-63c9-49eb-aa3d-5d19f6997312}")

#define PREF_TEXT_ANTIALIAS_MODE "antialiasing_mode"

#define PREF_GAMMA "gamma"
#define PREF_ENHANCED_CONTRAST "enhanced_contrast"
#define PREF_CLEAR_TYPE_LEVEL "cleartype_level"
#define PREF_PIXEL_GEOMETRY "pixel_geometry"
#define PREF_RENDERING_MODE "rendering_mode"

#define PREF_THRESHOLD "threshold"

struct Params {
  D2D1_TEXT_ANTIALIAS_MODE AntialiasMode;
  IDWriteRenderingParams *RenderingParams;

  FLOAT Gamma;
  FLOAT EnhancedContrast;
  FLOAT ClearTypeLevel;
  DWRITE_PIXEL_GEOMETRY PixelGeometry;
  DWRITE_RENDERING_MODE RenderingMode;
};

int Threshold = 16;
Params Small = {D2D1_TEXT_ANTIALIAS_MODE_DEFAULT, NULL,
                1.8f, 0.5f, 0.5f, DWRITE_PIXEL_GEOMETRY_RGB,
                DWRITE_RENDERING_MODE_DEFAULT};
Params Large = {D2D1_TEXT_ANTIALIAS_MODE_DEFAULT, NULL,
                1.8f, 0.5f, 0.5f, DWRITE_PIXEL_GEOMETRY_RGB,
                DWRITE_RENDERING_MODE_DEFAULT};
bool Initialized = FALSE;
HINSTANCE HInst = NULL;
TCHAR IniFile[MAX_PATH];

void Update();
INT GetIntPref(TCHAR *section, TCHAR *key);
void PrefsToParams(TCHAR *aPrefBranch, Params *aParams);
void Initialize();
void Finalize();

#ifdef DEBUG
#define DBG WriteDebugInfo
VOID WINAPI WriteDebugInfo(const char *format, ...)
{
  va_list ap;
  char buff[256];
  va_start(ap, format);
  vsprintf(buff, format, ap);
  va_end(ap);
  MessageBox(0, buff, _T("Aatie"), MB_OK);
  //  OutputDebugString(buff);
}
#else
#define DBG __noop
#endif

class Aatie : public IObjectWithSite
{
public:
  STDMETHODIMP QueryInterface(REFIID riid, void **ppvObject);
  STDMETHODIMP_(ULONG) AddRef();
  STDMETHODIMP_(ULONG) Release();
  
  STDMETHODIMP SetSite(IUnknown *pUnkSite);
  STDMETHODIMP GetSite(REFIID riid, void **ppvSite);

  Aatie();
  ~Aatie();

private:
  LONG _Ref;
};

//

class Detour : public ID2D1RenderTarget
{
public:
  static void (WINAPI Detour::* TrueDrawGlyphRun)(
    D2D1_POINT_2F baselineOrigin,
    const DWRITE_GLYPH_RUN *glyphRun,
    ID2D1Brush *foregroundBrush,
    DWRITE_MEASURING_MODE measuringMode);

  void WINAPI FakeDrawGlyphRun(
    D2D1_POINT_2F baselineOrigin,
    const DWRITE_GLYPH_RUN *glyphRun,
    ID2D1Brush *foregroundBrush,
    DWRITE_MEASURING_MODE measuringMode);
};

//

STDMETHODIMP Aatie::QueryInterface(REFIID riid, void **ppvObject)
{
  if (IsEqualIID(riid, IID_IUnknown)) {
    *ppvObject = static_cast<Aatie *>(this);
  } else if (IsEqualIID(riid, IID_IObjectWithSite)) {
    *ppvObject = static_cast<IObjectWithSite *>(this);
  } else {
    *ppvObject = NULL;
    return E_NOINTERFACE;
  }
  AddRef();
  return S_OK;
}

STDMETHODIMP_(ULONG) Aatie::AddRef()
{
  return InterlockedIncrement(&_Ref);
}

STDMETHODIMP_(ULONG) Aatie::Release()
{
  if (InterlockedDecrement(&_Ref)) {
    return _Ref;
  }
  delete this;
  return 0;
}

STDMETHODIMP Aatie::SetSite(IUnknown *pUnkSite)
{
  return S_OK;
}

STDMETHODIMP Aatie::GetSite(REFIID riid, void **ppvSite)
{
  return E_NOINTERFACE;
}

Aatie::Aatie() : _Ref(0)
{
}

Aatie::~Aatie()
{
}

STDAPI DllGetClassObject(REFCLSID rclsid, REFIID riid, LPVOID *ppvOut)
{
  Update();
  Initialize();
  return CLASS_E_CLASSNOTAVAILABLE;
}

STDAPI DllCanUnloadNow(void)
{ 
  return S_OK;
}

STDAPI DllRegisterServer(void)
{
  TCHAR dfile[MAX_PATH];
  GetModuleFileName(HInst, dfile, sizeof(dfile) / sizeof(TCHAR));

  SHSetValue(HKEY_CLASSES_ROOT,
             _T("CLSID\\") CLSID_Aatie_s,
             _T(""), REG_SZ,
             _T("Anti-Aliasing Tuner for IE"), 26 * sizeof(TCHAR));

  SHSetValue(HKEY_CLASSES_ROOT,
             _T("CLSID\\") CLSID_Aatie_s _T("\\InProcServer32"),
             _T(""), REG_SZ,
             dfile, ((int)_tcslen(dfile)+1) * sizeof(TCHAR));

  SHSetValue(HKEY_CLASSES_ROOT,
             _T("CLSID\\") CLSID_Aatie_s _T("\\InProcServer32"),
             _T("ThreadingModel"), REG_SZ,
             _T("Apartment"), 10 * sizeof(TCHAR));

  SHSetValue(HKEY_LOCAL_MACHINE,
             _T("SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Browser Helper Objects\\") CLSID_Aatie_s,
             _T(""), REG_SZ,
             _T("Anti-Aliasing Tuner for IE"), 26 * sizeof(TCHAR));

  DWORD dw = 1;
  SHSetValue(HKEY_LOCAL_MACHINE,
             _T("SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Browser Helper Objects\\") CLSID_Aatie_s,
             _T("NoExplorer"), REG_DWORD,
             &dw, sizeof(DWORD));

  return S_OK;
}

STDAPI DllUnregisterServer(void)
{
  SHDeleteKey(HKEY_CLASSES_ROOT, _T("CLSID\\") CLSID_Aatie_s);
  SHDeleteKey(HKEY_LOCAL_MACHINE, _T("SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Browser Helper Objects\\") CLSID_Aatie_s);
  return S_OK;
}

BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
  if (fdwReason == DLL_PROCESS_ATTACH) {
    DisableThreadLibraryCalls(hinstDLL);
    HInst = hinstDLL;
  } else if (fdwReason == DLL_PROCESS_DETACH) {
    Finalize();
  }
  return TRUE;
}

HMODULE WINAPI Detoured() {
  return HInst;
}


//////////////////////////////////////////////////////////////////////////////

void Update()
{
  GetModuleFileName(HInst, IniFile, sizeof(IniFile) / sizeof(TCHAR));
  _tcscpy_s(&IniFile[_tcslen(IniFile) - 3], 4, _T("ini"));

  PrefsToParams(_T("NormalFont"), &Small);
  PrefsToParams(_T("LargeFont"), &Large);
  Threshold = (int)GetIntPref("Aatie", PREF_THRESHOLD);

  return;
}

INT GetIntPref(TCHAR *section, TCHAR *key)
{
  TCHAR buf[32];
  GetPrivateProfileString(section, key, _T("-1"), buf, sizeof(buf), IniFile);
  if ((_T('0') <= buf[0] && buf[0] <= _T('9')) ||
      ((buf[0] == _T('-')) && (_T('0') <= buf[1] && buf[1] <= _T('9')))) {
    return atoi(buf);
  } else {
    return -1;
  }
}

void PrefsToParams(TCHAR *aPrefBranch, Params *aParams)
{
  CComPtr<IDWriteFactory> dw_factory;
  DWriteCreateFactory(DWRITE_FACTORY_TYPE_SHARED,
                      __uuidof(IDWriteFactory),
                      (IUnknown**)&dw_factory);
  CComPtr<IDWriteRenderingParams> dwrpm;
  dw_factory->CreateRenderingParams(&dwrpm);

  //
  INT tmp;
  tmp = GetIntPref(aPrefBranch, PREF_TEXT_ANTIALIAS_MODE);
  aParams->AntialiasMode = (D2D1_TEXT_ANTIALIAS_MODE)tmp;

  //
  tmp = GetIntPref(aPrefBranch, PREF_GAMMA);
  aParams->Gamma =
    (tmp < 0) ? dwrpm->GetGamma() : (FLOAT)tmp / 1000;

  tmp = GetIntPref(aPrefBranch, PREF_ENHANCED_CONTRAST);
  aParams->EnhancedContrast =
    (tmp < 0) ? dwrpm->GetEnhancedContrast() : (FLOAT)tmp / 100;

  tmp = GetIntPref(aPrefBranch, PREF_CLEAR_TYPE_LEVEL);
  aParams->ClearTypeLevel =
    (tmp < 0) ? dwrpm->GetClearTypeLevel() : (FLOAT)tmp / 100;

  tmp = GetIntPref(aPrefBranch, PREF_PIXEL_GEOMETRY);
  aParams->PixelGeometry =
    (tmp < 0) ? dwrpm->GetPixelGeometry() : (DWRITE_PIXEL_GEOMETRY)tmp;

  tmp = GetIntPref(aPrefBranch, PREF_RENDERING_MODE);
  aParams->RenderingMode = (DWRITE_RENDERING_MODE)tmp;

  //
  if (aParams->RenderingParams) {
    aParams->RenderingParams->Release();
  }

  dw_factory->CreateCustomRenderingParams(
    aParams->Gamma,
    aParams->EnhancedContrast,
    aParams->ClearTypeLevel,
    aParams->PixelGeometry,
    aParams->RenderingMode,
    &aParams->RenderingParams);

  DBG("PrefsToParams: %f %f %f %x %x %x\n",
      aParams->Gamma,
      aParams->EnhancedContrast,
      aParams->ClearTypeLevel,
      aParams->PixelGeometry,
      aParams->RenderingMode,
      aParams->AntialiasMode);
}

//////////////////////////////////////////////////////////////////////////////

void (WINAPI Detour::* Detour::TrueDrawGlyphRun)(
  D2D1_POINT_2F baselineOrigin,
  const DWRITE_GLYPH_RUN *glyphRun,
  ID2D1Brush *foregroundBrush,
  DWRITE_MEASURING_MODE measuringMode = DWRITE_MEASURING_MODE_NATURAL) = NULL;

void WINAPI Detour::FakeDrawGlyphRun(
  D2D1_POINT_2F baselineOrigin,
  const DWRITE_GLYPH_RUN *glyphRun,
  ID2D1Brush *foregroundBrush,
  DWRITE_MEASURING_MODE measuringMode = DWRITE_MEASURING_MODE_NATURAL)
{
  // DBG("#   FakeDrawGlyphRun: START fontEmSize=%f\n", glyphRun->fontEmSize);
  Params *font;
  font = (glyphRun->fontEmSize < Threshold) ? &Small : &Large;

  this->SetTextRenderingParams(font->RenderingParams);

  if (font->AntialiasMode < 0) {
  } else {
    D2D1_TEXT_ANTIALIAS_MODE a = this->GetTextAntialiasMode();
    D2D1_TEXT_ANTIALIAS_MODE b = font->AntialiasMode;
    if ((a == D2D1_TEXT_ANTIALIAS_MODE_GRAYSCALE) &&
        (b == D2D1_TEXT_ANTIALIAS_MODE_CLEARTYPE)) {
      b = D2D1_TEXT_ANTIALIAS_MODE_GRAYSCALE;
    }
    this->SetTextAntialiasMode(b);
  }

  (this->*TrueDrawGlyphRun)(baselineOrigin,
                            glyphRun,
                            foregroundBrush,
                            measuringMode);
  // DBG("#   FakeDrawGlyphRun: END\n");
}

//////////////////////////////////////////////////////////////////////////////

typedef HRESULT (WINAPI * __D2D1CreateFactory)(
  D2D1_FACTORY_TYPE factoryType,
  REFIID riid,
  const D2D1_FACTORY_OPTIONS *pFactoryOptions,
  void **ppIFactory);

void Initialize()
{
  if (Initialized) {
    return;
  }

  //
  __D2D1CreateFactory _D2D1CreateFactory = (__D2D1CreateFactory)
    GetProcAddress(LoadLibraryW(L"d2d1.dll"), "D2D1CreateFactory");
  if (!_D2D1CreateFactory) {
    return;
  }

  CComPtr<ID2D1Factory> d2d_factory;
  _D2D1CreateFactory(D2D1_FACTORY_TYPE_SINGLE_THREADED,
                     __uuidof(ID2D1Factory), NULL, (void**)&d2d_factory);

  //
  CComPtr<IWICImagingFactory> factory;
  factory.CoCreateInstance(CLSID_WICImagingFactory);
  CComPtr<IWICBitmap> bitmap;
  factory->CreateBitmap(32,
                        32,
                        GUID_WICPixelFormat32bppPBGRA,
                        WICBitmapCacheOnLoad,
                        &bitmap);
  const D2D1_PIXEL_FORMAT format =
    D2D1::PixelFormat(DXGI_FORMAT_B8G8R8A8_UNORM,
                      D2D1_ALPHA_MODE_PREMULTIPLIED);
  const D2D1_RENDER_TARGET_PROPERTIES properties =
    D2D1::RenderTargetProperties(D2D1_RENDER_TARGET_TYPE_DEFAULT,
                                 format,
                                 0.0f,
                                 0.0f,
                                 D2D1_RENDER_TARGET_USAGE_NONE);
  CComPtr<ID2D1RenderTarget>target;
  d2d_factory->CreateWicBitmapRenderTarget(bitmap,
                                           properties,
                                           &target);

  //
  Detour::TrueDrawGlyphRun = brute_cast<void(WINAPI Detour::*)(
    D2D1_POINT_2F,
    const DWRITE_GLYPH_RUN *,
    ID2D1Brush *,
    DWRITE_MEASURING_MODE)>(
      (*reinterpret_cast<void***>(target.p))[29]);

  DetourTransactionBegin();
  DetourUpdateThread(GetCurrentThread());
  DetourAttach(&(PVOID&)Detour::TrueDrawGlyphRun,
               (PVOID)(&(PVOID&)Detour::FakeDrawGlyphRun));
  DetourTransactionCommit();

  Initialized = TRUE;
  return;
}

void Finalize()
{
  if (!Initialized) {
    return;
  }

  DetourTransactionBegin();
  DetourUpdateThread(GetCurrentThread());
  DetourDetach(&(PVOID&)Detour::TrueDrawGlyphRun,
               (PVOID)(&(PVOID&)Detour::FakeDrawGlyphRun));
  DetourTransactionCommit();

  Initialized = FALSE;
  return;
}
